# -*- coding: utf-8 -*-
# Authors : Lucas Terriel <lucas.terriel@inria.fr>
#           Alix Chagué <alix.chague@inria.fr>
# Licence : MIT
"""
    The ``Prediction`` module to generate HTR/OCR pipeline on data
    ==============================================================

"""
import unicodedata

from kami.kamutils._utils import _report_log
from kami.transcription._base_prediction import (_binarize_image,
                                                 _segment_image,
                                                 _predict_transcription)

__all__ = [
    "_Prediction",
]

class _Prediction:
    """this class create a prediction from an offline model, image, and ground truth file.

    User cannot access to this class via :class: `Kami` facade class.
    """
    def __init__(self, filename, image, model_loaded, verbosity, code, reference=None):
        self.filename = filename
        self.image = image
        self.model = model_loaded
        self.verbosity = verbosity
        self.code = code
        self.reference = reference

    def get_transcription(self):
        """built from image/coordinates and an offline recognition model of the text/xml
        a transcript.

        Process Text
        ------------
        * 1- Binarization
        * 2- Segmentation
        * 3- Prediction (Text recognition)
        * 4- Convert transcription kraken object in string format

        Process XML
        ------------
        * 1- Extract Segments/Bounds
        * 2- Prediction (Text recognition)
        * 3- Convert transcription kraken object in string format
        """

        if self.verbosity:
            _report_log(
                f"{'#' * 10} HTR/OCR Pipeline initialize "
                f"for {self.code} file : {self.filename} ... "
                f"{'#' * 10}",
                "I"
            )

        canvas = ""

        try:
            if self.code == "txt":
                image_bin = _binarize_image(self.image, self.verbosity)
                image_seg = _segment_image(image_bin, self.verbosity)
                prediction = _predict_transcription(image=image_bin,
                                                    bound=image_seg,
                                                    model_loaded=self.model,
                                                    verbosity=self.verbosity,
                                                    code="txt")
                for line in prediction:
                    # .prediction is a kraken_ocr_record class attribute for recover the text in
                    # kraken.rpred.mm_rpred object
                    canvas += f"{unicodedata.normalize('NFC', line.prediction)}" + "\n"
                    if self.verbosity:
                        _report_log(
                            f"{'#' * 10} Kraken object "
                            f"in string format converted  {'#' * 10}",
                            "S"
                        )

            if self.code == "xml":
                for bound in self.reference.bounds:
                    prediction = _predict_transcription(image=self.image,
                                                        bound=bound,
                                                        model_loaded=self.model,
                                                        verbosity=self.verbosity,
                                                        code="xml")

                    canvas += f"{unicodedata.normalize('NFC', prediction)}" + "\n"
                    if self.verbosity:
                        _report_log(
                            f"{'#' * 10} Kraken object in "
                            f"string format converted  {'#' * 10}",
                            "S"
                        )

        except Exception as exception:
            _report_log(
                f"Error : unable to transcribe {self.code} file",
                "E"
            )
            _report_log(
                f"type : {exception}"
            )

        # run the model and get a
        # prediction -> https://github.com/mittagessen/kraken/blob/master/kraken/rpred.py#L353
        # return a list of transcription (can we have the same type of objects as
        # get_transcriptions_txt()
        return canvas

    def __str__(self):
        return f"Prediction(\n" \
               f"filename={self.filename},\n" \
               f"image={self.image},\n" \
               f"model={self.model}),\n" \
               f"verbosity={self.verbosity},\n" \
               f"code={self.code},\n" \
               f"reference={self.reference}"



# -*- coding: utf-8 -*-
# Authors : Lucas Terriel <lucas.terriel@inria.fr>
#           Alix Chagué <alix.chague@inria.fr>
# Licence : MIT
"""Common code for prediction process.
"""


from kami.kamutils._utils import _report_log
from kraken import (rpred,
                    pageseg,
                    binarization)

def _binarize_image(image_loaded: object, verbosity: bool) -> object:
    """Binarize a series of images"""
    try:
        # create binarized image
        im_bin = binarization.nlbin(image_loaded)
        if verbosity:
            _report_log(f"{'#' * 10} Binarization procceded  {'#' * 10}", "S")
    except Exception as exception:
        _report_log(f"type : {exception}")
        _report_log(f"Error : unable to binarize - {image_loaded}", "E")

    if verbosity:
        _report_log(im_bin, "V")

    return im_bin


def _segment_image(image_loaded: object, verbosity: bool) -> list:
    """Perform segmentation on a image"""
    try:
        segments_image = pageseg.segment(image_loaded, text_direction='horizontal-lr')
        if verbosity:
            _report_log(f"{'#' * 10} Segmentation procceded  {'#' * 10}", "S")
    except Exception as exception:
        _report_log(f"type : {exception}")
        _report_log(f"Error : unable to segment - {image_loaded}", "E")

    if verbosity:
        _report_log(segments_image, "V")

    return segments_image


def _predict_transcription(image: object, bound, model_loaded: object, verbosity: bool, code: str) -> str:
    """Perform transcription on an images given a segment"""

    # created the text prediction (kraken.rpred.mm_rpred object)
    # see https://github.com/mittagessen/kraken/issues/213
    generator = rpred.rpred(network=model_loaded,
                               im=image,
                               bounds=bound,
                               pad=16,
                               bidi_reordering=True)
    if code == "xml":
        nxt_gen = next(generator)
        text = nxt_gen.prediction

        if verbosity:
            _report_log(f"{'#' * 10} Text recognition procceded  {'#' * 10}", "S")
        if verbosity:
            _report_log(generator, "V")
        return text

    if code == "txt":
        return generator



# -*- coding: utf-8 -*-
# Authors : Lucas Terriel <lucas.terriel@inria.fr>
#           Alix Chagué <alix.chague@inria.fr>
# Licence : MIT
"""Common code for inputs in prediction process.
"""

import sys

from kraken import lib
from PIL import Image

from kami.kamutils._utils import _report_log

def _load_model(model: str, verbosity: bool) -> object:
    """Load an mlmodel with Kraken"""
    # STEP 1 : Loading model
    # Loading the model like kraken.lib.models.TorchSeqRecognizer object
    # --- Issue : find a way to ignore the output precautionary message
    model_loaded = lib.models.load_any(model)
    # At each new step, a validation message is displayed :
    if verbosity:
        _report_log(f"{'#' * 10} Model loaded  {'#' * 10}", "S")
        _report_log(model_loaded, "V")
    return model_loaded

def _load_image(image: str, verbosity: bool) -> list:
    """Load images with PIL and return a list of loaded images (PIL objects)"""
    # At each new step we create an empty list to receive the new objects, such as :
    try:
        img_pil = Image.open(image)
        if verbosity:
            _report_log(f"{'#' * 10} Image loaded  {'#' * 10}", "S")
    except Exception as exception:
        _report_log(f"type : {exception}")
        _report_log(f"Error : unable to load images - {image}", "E")
        sys.exit('program exit')

    if verbosity:
        _report_log(img_pil, "V")

    return img_pil



# -*- coding: utf-8 -*-
# Authors : Lucas Terriel <lucas.terriel@inria.fr>
# Licence : MIT
"""
    The ``evaluation`` module to assess text recognition (OCR/HTR) scores
    ======================================================================

"""

from typing import Sequence, Tuple, Union
import Levenshtein
from ._base_metrics import (_truncate_score,
                            _hot_encode,
                            _get_percent)

__all__ = [
    "Scorer",
]


class Scorer:
    """This class calculates the set of HTR / OCR scores.

    User can accesses directly to metrics/scores via the attributes of the :class: `Scorer` class
    in constructor method or via :class: `Kami` facade class.

    Parameters
    ----------

        :param reference: Ground-truth text or string
        :type reference: str
        :param prediction: model string or text prediction or test text or string
        to compare to :param: `reference` parameter
        :type prediction: str
        :param show_percent: `True` if the user want to show result in percent else `False`, defaults to False
        :type show_percent: bool, optional
        :param truncate_score: `True` if the user want to truncate result in percent else `False`,
        can indicate the number of digits after floating point with :param: `round_digits` parameter, defaults to False
        :type truncate_score: bool, optional
        :param round_digits: Set the number of digits after floating point in string form, defaults to '.01'
        :type round_digits: str, optional

    Attributes
    ----------

        :ivar _opt_percent: Option to show result in percent
        :type _opt_percent: bool
        :ivar _opt_truncate: Option to truncate result
        :type _opt_truncate: bool
        :ivar _round_digits: Number of digits after floating points
        :type _round_digits: str
        :ivar reference: Ground-truth text or string
        :type reference: str
        :ivar prediction: Text to compare or predicted sequence
        :type prediction: str
        :ivar length_char_reference: Total number of characters in reference string
        :type length_char_reference: int
        :ivar length_char_prediction: Total number of characters in predicted string
        :type length_char_prediction: int
        :ivar length_words_reference: Total number of words in reference string
        :type length_words_reference: int
        :ivar length_words_prediction: Total number of words in predicted string
        :type length_words_prediction: int
        :ivar lev_distance_words: Edit distance (levensthein) between reference and prediction based on words
        :type lev_distance_words: int
        :ivar lev_distance_char: Edit distance (levensthein) between reference and prediction based on characters
        :type lev_distance_char: int
        :ivar hamming: Hamming distance between reference and prediction based on words
        :type hamming: int or str ("Ø") if sequences have not the same numbers of characters
        :ivar wer: Word Error Rate
        :type wer: float
        :ivar cer: Character Error Rate
        :type cer: float
        :ivar wacc: Word Accuracy
        :type wacc: float
        :ivar cip: Character Information Preserved
        :type cip: float
        :ivar cil: Character Information Lost
        :type cil: float
        :ivar mer: Match Error Rate
        :type mer: float
        :ivar hits: number of characters that matches between reference and prediction string
        :type hits: int
        :ivar substs: number of characters substituted between reference and prediction string
        :type substs: int
        :ivar deletions: number of characters deleted between reference and prediction string
        :type deletions: int
        :ivar insertions: number of characters inserted between reference and prediction string
        :type insertions: int
        :ivar board: A benchmark of all metrics
        :type board: dict


    """
    def __init__(self,
                 reference: str,
                 prediction: str,
                 show_percent: bool = False,
                 truncate_score: bool = False,
                 round_digits: str = '.01'):

        # Options
        self._opt_percent = show_percent
        self._opt_truncate = truncate_score
        self._round_digits = round_digits

        # Strings to compare
        self.reference = reference
        self.prediction = prediction

        # Length set of sentences
        self.length_char_reference = len(reference)
        self.length_char_prediction = len(prediction)
        self.length_words_reference = len(reference.split())
        self.length_words_prediction = len(prediction.split())

        # Distances
        self.lev_distance_words, self.lev_distance_char = self._levensthein_distance()
        self.hamming = self._hamming_distance()

        # Operations
        self.hits, self.substs, self.deletions, self.insertions = self._get_operation_counts()

        # HTR/OCR Metrics
        self.wer = self._wer()
        self.cer = self._cer()
        self.wacc = self._wacc()
        self.cip = self._cip()
        self.cil = self._cil()
        self.mer = self._mer()

        # Summary of all metrics
        self.board = {
                "levensthein_distance_char": self.lev_distance_char,
                "levensthein_distance_words": self.lev_distance_words,
                "hamming_distance": self.hamming,
                "wer": self.wer,
                "cer": self.cer,
                "wacc": self.wacc,
                "mer": self.mer,
                "cil": self.cil,
                "cip": self.cip,
                "hits": self.hits,
                "substitutions": self.substs,
                "deletions": self.deletions,
                "insertions": self.insertions
        }

    # Collection of distance metrics #

    def _levensthein_distance(self) -> Tuple[int, int]:
        """Compute Levensthein distance from C extension module Python-Levensthein."""
        # Compute levensthein at word level
        lev_distance_words = Levenshtein.distance(*_hot_encode(
            [self.reference.split(), self.prediction.split()]))
        # Compute levensthein at char level
        lev_distance_char = Levenshtein.distance(self.reference, self.prediction)
        return lev_distance_words, lev_distance_char

    def _hamming_distance(self) -> Union[str, int]:
        """Compute Hamming distance from C extension module Python-Levensthein."""
        if self.length_char_reference != self.length_char_prediction:
            hamming_score = "Ø"
        else:
            hamming_score = Levenshtein.hamming(self.reference, self.prediction)
        return hamming_score

    # Collection of HTR/OCR metrics #

    def _wer(self) -> float:
        """Compute word error rate (WER)."""
        wer = (self.lev_distance_words/self.length_words_reference)
        if self._opt_percent:
            wer = _get_percent(wer)
        if self._opt_truncate:
            wer = _truncate_score(wer, self._round_digits)
        return wer

    def _cer(self) -> float:
        """Compute character error rate (CER)."""
        cer = (self.lev_distance_char/self.length_char_reference)
        if self._opt_percent:
            cer = _get_percent(cer)
        if self._opt_truncate:
            cer = _truncate_score(cer, self._round_digits)
        return cer

    def _wacc(self) -> float:
        """Compute word accuracy (Wacc)."""
        wacc = (1 - (self.lev_distance_words/self.length_words_reference))
        if self._opt_percent:
            wacc = _get_percent(wacc)
        if self._opt_truncate:
            wacc = _truncate_score(wacc, self._round_digits)
        return wacc

    def _cip(self) -> Union[float, int]:
        """Compute character information preserved (CIP)."""
        if self.prediction:
            cip = (float(self.hits)/self.length_char_reference)*(float(self.hits)/self.length_char_prediction)
            if self._opt_percent:
                cip = _get_percent(cip)
            if self._opt_truncate:
                cip = _truncate_score(cip, self._round_digits)
        else:
            cip = 0
        return cip

    def _cil(self) -> Union[float, int]:
        """Compute character information lost (CIL)."""
        if self.prediction:
            cil = (1 - (float(self.hits)/self.length_char_reference)*(float(self.hits)/self.length_char_prediction))
        else:
            cil = 0
        if self._opt_percent:
            cil = _get_percent(cil)
        if self._opt_truncate:
            cil = _truncate_score(cil, self._round_digits)
        return cil

    def _mer(self) -> float:
        """Compute match error rate (MER)."""
        mer = float(
            self.substs
            + self.deletions
            + self.insertions) \
            / float(
            self.hits
            + self.substs
            + self.deletions
            + self.insertions)

        if self._opt_percent:
            mer = _get_percent(mer)
        if self._opt_truncate:
            mer = _truncate_score(mer, self._round_digits)
        return mer

    def _get_operation_counts(self) -> Tuple[int, int, int, int]:
        """Find sequence of edit operations transforming one string to another.
        Based on editops function from C extension module python-Levenshtein."""
        result_editops = Levenshtein.editops(self.reference, self.prediction)

        def _sum_operations(keyword: str, results: Sequence[Tuple[str, int, int]]) -> int:
            """Compute a sum of define operations"""
            total = sum(1 if operations[0] == keyword else 0 for operations in results)
            return total

        substitutions = _sum_operations("replace", result_editops)
        deletions = _sum_operations("delete", result_editops)
        insertions = _sum_operations("insert", result_editops)
        hits = self.length_char_reference - (substitutions + deletions)
        return hits, substitutions, deletions, insertions



# -*- coding : utf-8 -*-

from os import path
import io
import setuptools

import kami

here = path.abspath(path.dirname(__file__))

try:
    with io.open(path.join(here, 'README.rst'), encoding='utf-8') as f:
        long_description = f.read()
except:
    long_description = u"Python package focused on HTR / OCR models evaluation and based on the Kraken transcription system."


with open("requirements.txt", encoding="utf-8") as f:
    install_requires = f.read().splitlines()


CLASSIFIERS = [
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]


setuptools.setup(
    name=kami.__title__,
    version=kami.__version__,
    author="Lucas Terriel, Alix Chagué",
    author_email="lucas.terriel@inria.fr, alix.chague@inria.fr",
    license=kami.__licence__,
    description="Python package focused on HTR / OCR models evaluation and based on the Kraken transcription system.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://gitlab.inria.fr/dh-projects/kami/kami-lib",
    project_urls={
        "Bug Tracker": "https://gitlab.inria.fr/dh-projects/kami/kami-lib/-/issues/new?issuable_template=bug_report",
    },
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    classifiers=CLASSIFIERS,
    keywords=["HTR", "OCR", "Evaluation framework", "metrics", "handwritten text recognition", "optical character recognition"],
)